=begin
	Ruby X11 Client Library
	Copyright 2001 by Mathieu Bouchard

	X11/Display.rb: main file; display, screen, global requests.

	$Id: Display.rb,v 1.94 2003/08/11 20:08:47 matju Exp $

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

=end

require "socket"

require "X11/Base"
require "X11/XID"
require "X11/Event"
require "X11/Remote"

#class Object
#	alias old_p p
#	alias old_puts puts
#	def p(x)    print "[#{caller[0]}] "; old_p(x) end
#	def puts(x) print "[#{caller[0]}] "; old_puts(x) end
#end

module X11 # The whole file goes in it
#----------------------------------------------------------------#

ProtocolVersion = [11,0]
ProtocolMajor,ProtocolMinor = ProtocolVersion
DEFAULT_XFONT = \
	"-adobe-helvetica-medium-r-normal--14-*-*-*-c-*-iso8859-1,"\
	"-misc-fixed-medium-r-normal--14-*-*-*-c-*-jisx0208.1983-0"
DEFAULT_NFONT = 2

# because of some I/O bugs in nearly all versions of Ruby

#----------------------------------------------------------------#
# Forms for Handshake

class VisualInfo < Form; fields_are \
	[:visual_id,Visual],
	[:qlass,VisualClass],
	[:bits_per_rgb_value,Uint8],
	[:colormap_entries,Uint16],
	[:red_mask,  Uint32],
	[:green_mask,Uint32],
	[:blue_mask, Uint32], 4
end

class DepthInfo < Form; fields_are \
	[:depth,Uint8], 1,
	[:visuals,Uint16,:length], 4,
	[:visuals,List.of(VisualInfo),:data]
end

class ScreenInfo < Form; fields_are \
	[:root,Window],
	[:default_colormap,Colormap],
	[:white_pixel,Colornum],
	[:black_pixel,Colornum],
	[:current_input_masks,EventMask],
	[:size_in_pixels,     Size],
	[:size_in_millimeters,Size],
	[:min_installed_maps,Uint16],
	[:max_installed_maps,Uint16],
	[:root_visual,Visual],
	[:backing_stores,BackingStore],
	[:save_unders,Bool],
	[:root_depth,Uint8],
	[:depths,Uint8,:length],
	[:depths,List.of(DepthInfo),:data]

	# attr_reader :display
end

class FormatInfo < Form; fields_are \
	[:depth,Uint8],
	[:bits_per_pixel,Uint8],
	[:scanline_pad,Uint8], 5
end

class DisplayInfo < Form; fields_are \
	[:release_number,Uint32],
	[:resource_id_base,Uint32],
	[:resource_id_mask,Uint32],
	[:motion_buffer_size,Uint32],
	[:vendor,Uint16,:length],
	[:maximum_request_length,Uint16],
	[:screens,Uint8,:length],
	[:formats,Uint8,:length],
	[:image_byte_order,Significance],
	[:bitmap_bit_order,Significance],
	[:bitmap_format_scanline_unit,Uint8],
	[:bitmap_format_scanline_pad,Uint8],
	[:min_keycode,KeyCode],
	[:max_keycode,KeyCode], 4,
	[:vendor,String8,:data],
	[:formats,List.of(FormatInfo),:data],
	[:screens,List.of(ScreenInfo),:data]
end

#----------------------------------------------------------------#

class ClientHandshake < Form; fields_are \
	[:byte_order, ByteOrder], 1,
	[:proto_version, List.of(Uint16,2)],
	[:auth_proto_name, Uint16, :length],
	[:auth_proto_data, Uint16, :length], 2,
	[:auth_proto_name, String8, :data],
	[:auth_proto_data, String8, :data]
end

class ServerHandshakeFailed < Form; fields_are \
	[:reason,Uint8,:length],
	[:version,List.of(Uint16,2)],
	[:xlen, Uint16],
	[:reason,String8Unpadded,:data]
end

class ServerHandshakeSuccess < Form; fields_are \
	1, [:version,List.of(Uint16,2)],
	[:xlen, Uint16]
end

class ServerHandshakeAuthenticate < Form; fields_are \
	5, [:xlen,Uint16]
#	[:reason,String8]
end

#----------------------------------------------------------------#
# miscellaneous structures used for Display's requests.

class KeyboardControl < PartialForm; fields_are \
	[:key_click_volume, Int8], # percent
	[:bell_volume, Int8], # percent
	[:bell_pitch, Int16],
	[:bell_duration, Int16],
	[:led, Uint8],
	[:led_mode, OnOff],
	[:key, KeyCode],
	[:auto_repeat_mode, OnOffDefault]
end

class PointerControl < Form; fields_are \
	[:numerator, Uint16],
	[:denominator, Uint16],
	[:threshold, Uint16]
end

class ScreenSaverControl < Form; fields_are \
	[:timeout, Uint16],
	[:interval, Uint16],
	[:prefer_blanking,ScreenSaver],
	[:allow_exposures,ScreenSaver]
end

#----------------------------------------------------------------#
class Display; extend RemoteClass; include RemoteObject

# category: attributes

	attr_reader :sock, :host
	attr_reader :display_id, :screen_id
	attr_reader :display_info, :format_info
	attr_reader :auth_proto_name, :auth_proto_data

	attr_reader :event_queue
	attr_reader :reply_set #!@#$ incl errors ???
	attr_reader :waiting_for

	attr_accessor :pause_on_error

	def screens; display_info.screens; end

	def initialize(target=nil,auth_info=nil)
		@xdisplay = self
		@event_queue = []
		@reply_set = {}
		@atom_cache = {}
		@auth_info = auth_info

		connect_to(target)
	end

# category: misc

	def inspect
		sprintf "\#<#{self.class} #{host}:#{display_id}.#{screen_id}>"
	end

# category: handshake

	def connect_to(target)
		target ||= ::ENV["DISPLAY"]
		@host,@display_id,@screen_id = nil
		case target
		when Socket
			@sock = target
		when String
			# /^(?:[^:]*?\/)?(.*):(\d+)(?:.(\d+))?$/
			md = /^([\w.-]*):(\d+)(?:.(\d+))?$/.match(target)
			@host,@display_id,@screen_id = md[1],md[2],md[3]
			# @host = "" if @host.length==0
			if @host == "" then
				@sock = UNIXSocket.new("/tmp/.X11-unix/X#{@display_id}")
				# @host = "localhost"
				@family = :Local
			else
				@sock = TCPSocket.new(@host,6000+@display_id.to_i)
				@family = :Internet
			end
			sock.extend NonBlockingSocket
			sock.extend X11::SocketLog if X11_LOG >= 0
		else
			raise "Please specify DISPLAY"
		end

		if X11_LOG >= 2; sock.log = true; end
		if X11_LOG >= 3; self.pause_on_error = true; end
		
		authorization
	end

	def authorization
		@auth_proto_name,@auth_proto_data = case @auth_info
			when Array; @auth_info
			else X11::Auth.new.get_by_host(@host||"localhost",@family,@display_id)
			end

		w = XStringWriter.new(self)
		ClientHandshake.xwrite(w,ClientHandshake.new(
			OurByteOrder,
			ProtocolVersion,
			@auth_proto_name,
			@auth_proto_data))
		sock.write w.chunk

		#[:answer,Uint8],
		ret = ServerHandshake.xread(sock)

		case ret
		when :Failed
			reason,version,xlen =
				ServerHandshakeFailed.xread(XIOReader.new(self,sock))
			raise "Connection to server failed " \
				"(version #{version.join'.'} -- #{reason}"
		when :Success
			version,xlen =
				ServerHandshakeSuccess.xread(XIOReader.new(self,sock))

			data = XStringReader.new(self,sock.read(xlen*4))
			@display_info = DisplayInfo.xread(data)

			@xid_next  = 0
			@xid_shift = 0
			while (@display_info.resource_id_mask >> @xid_shift) & 1 == 0
				@xid_shift += 1
			end

			@sid = 1

			sock.nonblock = true
			sock.sync = false
#			sock.sync = true

		when :Authenticate
			xlen, =
				ServerHandshakeAuthenticate.xread(XIOReader.new(self,sock))
			reason = sock.read(xlen*4)
			reason.sub! /\0+$/, "" # i'm not supposed to do this
			raise "authentication required -- #{reason}"
		else
			raise "unknown response"
		end
	end

# category: Requests, Replies, Events, Errors

	def pick_sid
		@sid += 1
		@sid - 1
	end

	def xreceive
		sock.flush
		chunk = sock.read(32)
		eid, = chunk.unpack("C")
		case eid
		when 0 # error
			error = XError.new(chunk) #!@#$ should be xread
			X11.log :error, error.inspect
			STDIN.readline if pause_on_error

		when 1 # reply
			header = chunk.slice! 0,8
			sid,extra = header.unpack("xxSL")
			#!@#$ BUG: when non-blocking, large replies may break.
			chunk << sock.read(4*extra) if extra>0
			X11.log1("Reply 0x%04x\n" % sid)
			reply_handler = @reply_set[sid]
			raise "Unexpected Reply!!!" if not reply_handler
			@reply_set.delete sid
			reader = XStringReader.new(self,chunk,header)
			reader.header.read(1) # drop one byte (to get to in_header field)
			reply_handler.call reader

		else #event
			event_type = EventTypeNumber[eid & 0x7f]
			if not event_type then
				raise "Unknown event type: #{eid.inspect}"
			end
			event = event_type.xread(XStringReader.new(self,chunk))
			# X11.log :event, event.inspect
			@event_queue << event
		end
	end

	def xreceive_blocking
		#!@#$ this is not reentrant (does it need to?)
		sock.nonblock = false
		begin xreceive ensure
			sock.nonblock = true
		end
	end

	def xreceive_many
		begin
			loop{xreceive}
		rescue Errno::EWOULDBLOCK
		end
	end

	def next_event
		@event_queue.shift
	end

	# makes sure all send/reply pairs initiated during the callback
	# are completed when it returns. (except for exceptions...)
	def wait_for(&proc)
		old_waiting_for = @waiting_for
		@waiting_for = []
		yield
		while @waiting_for.length > 0
			if @reply_set[@waiting_for[0]]
				xreceive_blocking
			else
				@waiting_for.shift
			end
		end
	ensure
		@waiting_for = old_waiting_for
	end

# category: connection management

	def flush; begin sock.flush; rescue Errno::EWOULDBLOCK; end; end

	def sync
		raise
		# XSync(disp->xdisplay, False);
	end

# category: other

	def new_id
		id = @xid_next
		@xid_next+=1
		
		((id << @xid_shift) &
			@display_info.resource_id_mask) |
				@display_info.resource_id_base
	end

# methods for: atom cache

	def atom(name); atoms(name)[0]; end

	def atoms(*names)
		atoms = {}
		wait_for {
			names.each {|name|
				if @atom_cache[name] then
					atoms[name] = @atom_cache[name]
				else
					self.intern_atom(name) {|atom,|
						atom.instance_eval{@name = name}
						atoms[name] = atom
						@atom_cache[name] ||= atom
					}
				end
			}
		}
		names.map {|k| atoms[k] }
	end

# methods for: rpc: atoms

	def_remote_with_reply :intern_atom, 16, [
		[:name, Uint16, :length], 2,
		[:name, String8, :data],
		[:only_if_exists, Bool, {:default => false}, nil, true],
	], [
		[:atom, Atom]]
	
# methods for: rpc: selection

	def_remote :set_selection_owner, 22, [
		[:owner, Any.of(Window,None)],
		[:selection, Atom],
		[:time, TimeOrNow]]

	def_remote_with_reply :get_selection_owner, 23, [
		[:selection, Atom],
	], [
		[:owner, Any.of(Window,None)]]

	def_remote :convert_selection, 24, [
		[:requestor, Window],
		[:selection, Atom],
		[:target, Atom],
		[:property, Any.of(Atom,None)],
		[:time, TimeOrNow]]

# methods for: rpc: grabs

	def_remote :ungrab_pointer, 27, [
		[:time, TimeOrNow]]

	def_remote :ungrab_button, 29, [
		[:button,Any.of(Button,AnyButton)],
		[:grab_window, Window],
		[:modifiers, Any.of(KeyMask,AnyModifier)]]

	def_remote :change_active_pointer_grab, 30, [
		[:cursor, Any.of(Cursor,None)],
		[:time, TimeOrNow],
		[:event_mask, PointerEventMask]]
		
	def_remote :ungrab_keyboard, 32, [
		[:time, TimeOrNow]]

	def_remote :allow_events, 35, [
		[:mode, AllowEventsMode, :in_header],
		[:time, TimeOrNow]]

	def_remote   :grab_server, 36, []
	def_remote :ungrab_server, 37, []

# methods for: rpc: misc 1

	def_remote :send_event, 25, [
		[:propagate, Bool, :in_header],
		[:destination, Any.of(Window,EventDestination)],
		[:event_mask, EventMask],
		[:event, Event]]

	def_remote :warp_pointer, 41, [
		[:from_window, Any.of(Window,None)],
		[:to_window,   Any.of(Window,None)],
		[:from_point, Point],
		[:size, Size],
		[:to_point,   Point]]

	def_remote :set_input_focus, 42, [
		[:focus, Any.of(Window,SetInputFocus)],
		[:revert_to, InputFocusRevertTo, :in_header],
		[:time, TimeOrNow]]

	def_remote_with_reply :get_input_focus, 43, [
	], [
		[:revert_to, InputFocusRevertTo, :in_header],
		[:focus, Any.of(Window,SetInputFocus)]]

# methods for: rpc: fonts

	def_remote_with_reply :list_fonts, 49, [
		[:max_names, Uint16],
		[:pattern, Uint16, :length],
		[:pattern, String8, :data],
	], [
		[:names, Uint16, :length], 22,
		[:names, List.of(Str), :data]]

	def_remote_with_reply :list_fonts_with_info, 50, [
		[:max_names, Uint16],
		[:pattern, Uint16, :length],
		[:pattern, String8, :data],
	], [
		#!@#$ any number of replies like this one:
		[:name, Uint8, :length_in_header],

		#!@#$ the next 12 fields could become FontInfo
		#!@#$ except for :properties.
		[:min_bounds, CharInfo], 4,
		[:max_bounds, CharInfo], 4,
		[:min_char_or_byte2, Uint16],
		[:max_char_or_byte2, Uint16],
		[:default_char, Uint16],
		[:properties, Uint16, :length],
		[:draw_direction, DrawDirection],
		[:min_byte1, Uint8],
		[:max_byte1, Uint8],
		[:all_chars_exist, Bool],
		[:font_ascent,  Uint16],
		[:font_descent, Uint16],

		[:replies_hint, Uint32],
		[:properties, List.of(FontProp), :data],
		[:name, String8, :data]]

	def_remote :set_font_path, 51, [
		[:path, Uint16, :length], 2,
		[:path, List.of(Str), :data]]

	def_remote_with_reply :get_font_path, 52, [
	], [
		[:path, Uint16, :length], 22,
		[:path, List.of(Str), :data]]

# methods for: rpc: extensions

	def_remote_with_reply :query_extension, 98, [
		[:name, Uint16, :length], 2,
		[:name, String8, :data],
	], [
		[:present, Bool],
		[:major_opcode, Uint8],
		[:first_event, Uint8],
		[:first_error, Uint8]]

	def_remote_with_reply :list_extensions, 99, [
	], [
		[:names, Uint8, :length_in_header], 24,
		[:names, List.of(Str), :data]]

# methods for: rpc: keyboard config

	def_remote_with_reply :query_keymap, 44, [
	], [
		[:keys, List.of(Uint8,32)]]
	
	def_remote :change_keyboard_mapping, 100, [
		[:first_keycode, KeyCode],
		[:keysyms_per_keycode, Uint8], 2,
		[:keysyms, List.of(Keysym)], # count * keysyms_per_keycode
		[:keycode_count, Uint8, :in_header]]

	def_remote_with_reply :get_keyboard_mapping, 101, [
		[:first_keycode, KeyCode],
		[:count, Uint8],
	], [
		[:keysyms_per_keycode, Uint8, :in_header], 24,
		[:keysyms, List.of(Keysym)]] # count * keysyms_per_keycode

	def_remote :change_keyboard_control, 102, [
		[:control, KeyboardControl]]

	#!@#$ XKeyboardControl

	def_remote_with_reply :get_keyboard_control, 103, [
	], [
		[:global_auto_repeat, Bool, :in_header],
		[:led_mask, Uint32],
		[:key_click_volume, Int8], # percent
		[:bell_volume, Int8], # percent
		[:bell_pitch, Int16],
		[:bell_duration, Int16], 2,
		[:auto_repeats, List.of(Uint8,32)]]

# methods for: rpc: mouse config

	def_remote :change_pointer_control, 105, [
		[:control, PointerControl],
		[:do_acceleration, Bool],
		[:do_threshold,    Bool]]

	def_remote_with_reply :get_pointer_control, 106, [
	], [
		[:control, PointerControl]]

	def_remote_with_reply :set_pointer_mapping, 116, [
		[:map, Uint8, :length_in_header],
		[:map, List.of(Uint8), :data],
	], [
		[:status, MappingChangeStatus, :in_header]]

	def_remote_with_reply :get_pointer_mapping, 117, [
	], [
		[:map, Uint8, :length_in_header], 24,
		[:map, List.of(Uint8), :data]]

	def_remote_with_reply :set_modifier_mapping, 118, [
		[:keycodes_per_modifier, Uint8, :in_header],
		[:keycodes, List.of(KeyCode)], #!@#$ :keycodes_per_modifier * 8
	], [
		[:status, MappingChangeStatus, :in_header]]

	def_remote_with_reply :get_modifier_mapping, 119, [
	], [
		[:keycodes_per_modifier, Uint8, :in_header], 24,
		[:keycodes, List.of(KeyCode)]] #!@#$ :keycodes_per_modifier * 8

# methods for: rpc: screen saver

	def_remote :set_screen_saver, 107, [
		[:control, ScreenSaverControl]]

	def_remote_with_reply :get_screen_saver, 108, [
	], [
		[:control, ScreenSaverControl]]

	def_remote :force_screen_saver, 115, [
		[:mode, ScreenSaverAction, :in_header]]

# methods for: rpc: display admin

	def_remote :change_hosts, 109, [
		[:mode, ChangeMode, :in_header],
		[:family, HostFamily],
		[:address, Uint16, :length],
		[:address, List.of(Uint8), :data]]

	def_remote_with_reply :list_hosts, 110, [
	], [
		[:mode, AccessMode, :in_header],
		[:hosts, Uint16, :length], 22,
		[:hosts, List.of(Host), :data]]

	def_remote :set_access_control,  111, [[:mode, AccessMode,    :in_header]]
	def_remote :set_close_down_mode, 112, [[:mode, CloseDownMode, :in_header]]
	def_remote :kill_client, 113, [[:resource, Any.of(XID,AllTemporary)]]

# methods for: rpc: misc 2

	# broken
	def_remote :bell, 104, [
		[:volume, Uint8, {:in_header => true, :default => 100}]] # percent

	def_remote :no_operation, 127, []

end

#----------------------------------------------------------------#
end # of module X11

