=begin
	Ruby X11 Client Library
	Copyright 2001 by Mathieu Bouchard

	X11/Marshal.rb: X11-style Marshalling

	$Id: Marshal.rb,v 1.15 2003/08/11 20:08:06 matju Exp $

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

=end

require "X11/Type"

#----------------------------------------------------------------#
# This is for easier debugging only!!!

# pragdave's solution
# plus my vt-100 colour
module Kernel; def _(&b)
	f=yield
	STDERR.puts "\e[0;32m#{f}: \e[1m#{eval(f,b).inspect}\e[0m"
end; end

#----------------------------------------------------------------#

# sometimes the parameters of a method
# are insufficient to know the result.
class ItDependsError < RuntimeError; end

class Object
	def assert_type(name,value,type)
		return if type===value
		raise TypeError,
			"%s is of type %s, not %s" % [name,value.type,self.class.to_s],
			caller
	end
end

# for use by Keysyms.rb
module MAssignForHash
	def []=(*ks)
		vso = ks.pop
		vs = vso.to_a
		ks = ks[0] if Array===ks[0] # oops
		ks.length.times {|i| super(ks[i],vs[i]) }
		vso
	end
end

module X11
module Type; include Nameable; end # compatibility
#----------------------------------------------------------------#
# XStringReader, XStringWriter : string-buffers for use by marshallers

# note: from now on, there may be a lot of forward-references to Display.

class XStringReader
	attr_reader :xdisplay, :pos, :chunk

	def initialize(xdisplay,chunk,header=nil)
		@xdisplay,@chunk,@header = xdisplay,chunk,header
		@header &&= XStringReader.new(xdisplay,@header)
		#raise if not Display===@xdisplay
		@pos = 0
	end

	def header; @header or raise "no header here"; end
	def read(n); @pos+=n; @chunk[@pos-n,n]; end
	def pad; @pos += (4 - @pos%4) % 4; end
end

class XIOReader
	attr_reader :xdisplay, :pos, :io

	def initialize(xdisplay,io,header=nil)
		@xdisplay,@io,@header = xdisplay,io,header
		@header &&= XStringReader.new(xdisplay,@header)
		#raise if not Display===@xdisplay
		@pos = 0
	end

	def header; @header or raise "no header here"; end
	def read(n); @pos+=n; io.read(n); end
	def pad; read((4-@pos%4) % 4); end
end

class XStringWriter
	attr_reader :xdisplay, :chunk
	alias string chunk

	def initialize(xdisplay)
		@xdisplay = xdisplay
		@chunk = ""
	end

	def write(data); @chunk << data; end
	def header; @header ||= self.class.new(xdisplay); end
	def pad; @chunk << "\0"*((4 - @chunk.length%4) % 4); end
end

#----------------------------------------------------------------#
# adding marshal-specific typing to MetaRuby's built-in types.
# the marshalling itself is added later.

class <<Any
	alias [] new
	alias of new
end

class List
	alias initialize_old initialize
	ListTemplateOption = ChoiceType.new(:padded,:unpadded)
	#!@#$ wasn't the default supposed to be :padded ? check again
	def initialize(element_type,length=nil,options=:unpadded)
		assert_type "options",options,ListTemplateOption if DoCheckTypes
		initialize_old(element_type,length,options)
	end
end

class FormBase
	class<<self
		attr_reader :format
	end
end

class<<Form # metaclass
	alias old_fields_are fields_are

	# increments Type.rb's one with in-header and length options/semantics
	def fields_are(*fs)
		raise "Redefinition of a Form class' fields" if @fields
		f=nil

		@format = fs.map {|f|
			case f
			when Array
				if Symbol===f[2]
					f[4]=false
					case f[2]
					when :in_header; f[4]=true
					when :length; f[3]=:length
					when :data; f[3]=:data
					when :length_in_header; f[3]=:length; f[4]=true
					else raise "bad options=#{f[2].inspect}" end
					f[2]=nil
				end
				Slot[*f]
			when Integer; f=Slot[nil,Unused[f]]
			when Slot; f
			else raise "got #{f.inspect}, expected Array or Slot"
			end
		}

#		p @format
		foo = {} # for length/data
		(fields = [nil]).pop
		@format.each_with_index {|f,i|
			next if f.sym==nil or f.sym==:self
			if f.is_data
				j=foo[f.sym]
				raise "field not found for :data" if not j
				# can't always dup options
				fields[j] = Slot[f.sym,f.vtype,f.options]
				fields[j].part = nil
				@format[i].data_id = j
			elsif f.is_length
				foo[f.sym]=fields.length
				fields << f
			else
				fields << f
			end
		}
		old_fields_are *fields
		#compile_fields
	end
end

# don't use directly!
IN_HEADER = 1<<0
LENGTH    = 1<<1
DATA      = 1<<2

# although X11::Slot is a X11::Form, it's not meaningful to the
# X11 Server, so it is never encoded.

class Slot
	attr_accessor :data_id, :part

	alias initialize_old initialize
	def initialize(sym,vtype,options=nil,part=nil,in_header=false)
		Slot.check_types(sym,vtype,options) if DoCheckTypes
		@flags = 0
		#options[:default ] # this option has no default value (sic)
		initialize_old(sym,vtype,options)
		self.part = part
		self.in_header = (options && options[:in_header]) || in_header
#		@form[4] = in_header
		@flags |= IN_HEADER if in_header
		@flags |= case part when :length; LENGTH when :data; DATA; else 0 end
	end

	# in_header is defined twice. it shouldn't matter, but it's not nice.

	def part; @form[3]; end
	def part=x; @form[3]=x; end
	def in_header; @form[4]; end
	def in_header=x; @form[4]=x; end
#	def in_header ; @flags[0]>0; end
	def is_length ; @flags[1]>0; end
	def is_data   ; @flags[2]>0; end
	def is_self   ; @form[0]==:self; end
	def is_regular; @flags&0b110<1; end

	# this overrides (_replaces_) the stuff defined in Type.rb
	# takes into account the options and @format.

	valid_options = Hash # temporarily
	@format = @fields = [
		Slot[:sym,Any.of(Symbol,NilClass)],
			# nil = unused (fixed padding)
			# :self = the request's receiver.
			# other = normal slot
		Slot[:vtype,Type],
		Slot[:options,Any.of(Hash,NilClass)], # default nil
		Slot[:part,ChoiceType.of(nil,:length,:data)], # default nil
		Slot[:in_header,Boolean], # default false (nil, whatever)
	]
end

#----------------------------------------------------------------#
# PartialForm

class PartialForm < FormBase; extend Type

	class<<self; attr_reader :fields_h; end

	def self.fields_are(*fs)
		raise "Redefinition of a PartialForm class' fields" if @format
		f=nil

		@fields = @format = fs.map {|f|
			case f
			when Array; Slot[*f]
			when Slot; f
			when Integer; f=Slot[nil,Unused[f]]
			else raise "got #{f.inspect}, expected Array or Slot"
			end
		}

		@fields_h = {}
		@fields.each_with_index {|f,i| @fields_h[f.sym] = i }
	end

	def self.compile_fields
		return false if @compiled_fields
		e=""
		fields.each {|f| e << "def #{f.sym};@set[:#{f.sym}]end;" }
		class_eval e
		@compiled_fields = true
	end

	def self.check_types(h)
		for k,v in h
			i = fields_h[k]
			f = fields[i]
			if not f.vtype === v then
				raise TypeError, "for an #{self}, "\
					"supplied #{k}=#{v.inspect} "\
					"is not in type #{f.vtype.inspect}"
			end
		end
	end

# category: instance methods

	attr_reader :set

	def form
		t=Array.new(type.fields.length)
		for k,v in @set; t[type.fields_h[k]]=v; end
		t
	end
			
	def [](i); @form[i]; end

	def initialize(h={})
		self.class.check_types(h) if DoCheckTypes
		@set=h.dup
	end

	def ==(other); type == other.type && form == other.form; end
	alias eql? ==

	def hash; form.hash; end
	alias to_a   form
	alias to_ary form
	def to_h; @set.dup; end
end

#----------------------------------------------------------------#
# IntType (not in MetaRuby yet)

class IntType; include Type
	def initialize \
		( range, packer, length)
		 @range,@packer,@length =
		  range, packer, length
	end
	attr_reader :range,:packer,:length

	def ===(obj)
		Integer === obj && @range === obj && (!@proc || @proc[obj])
	end
	def such_that(&proc)
		r=dup
		r.instance_eval {
			@proc = proc
			# can't do that, actually
			#@str = ".such_that {...}"
		}
		r
	end
	def to_s
		super + @str.to_s
	end
end

#----------------------------------------------------------------#
# Unused is the type for performing padding
# use this as a metatype now
#!@#$ does not cache its instances!
class Unused; include Type
	attr_reader :bytes
	def initialize(bytes); @bytes = bytes; end
	class<<self; alias [] new; end

	def xread(input); input.read(bytes); nil; end

	def xwrite(output,object)
		s="\0"*bytes
		bytes.times{|i| s[i] = rand 256} if X11_LOG >= 4
		output.write s
	end
	def to_s; "#{self.class}[#{bytes}]"; end
	alias inspect to_s
end

#----------------------------------------------------------------#
# this is where the marshalling really begins
#----------------------------------------------------------------#

class Any
	# @xlength ||= @union.map {|t| t.xlength }.max

	def xwrite(output,object)
		good = @alternatives.find {|t| t === object }
		if not good then
			raise "#{object.inspect} is not in #{@union.join ','} of #{self}"
		end
		good.xwrite(output,object)
		output.pad
	end

	#!@#$ not right
	def xread(input)
		@alternatives[0].xread(input)
	end
end

#FormXWriteUsage = {}

class<<Form # metaclass
	def xwrite(output,object,receiver=nil)
		#X11::FormXWriteUsage[self]||=0
		#X11::FormXWriteUsage[self]+=1
		xwrite_compile if not @compiled
		if receiver
			then xwrite_compiled(output,object,receiver)
			else xwrite_compiled(output,object) end
	end

	# by the way, most optimization tricks used here
	# are completely worthless
	def xwrite_compile
#		@const=[]; iconst={}
		const=proc {|c|
#			if not iconst[c] then iconst[c]=@const.length; @const<<c end
#			"@const[#{iconst[c]}]"
			"#{c}"
		}
		code="def self.xwrite_compiled(output,form,receiver=nil)"
		code<<%{assert_type "form", form, self;} if DoCheckTypes
		i=0
		format.each {|f|
			target="output"; target<<".header" if f.in_header
			t=f.vtype

			vc = if f.is_self then "receiver"
				elsif not f.sym then "nil"
				elsif not f.is_data; i+=1; "form[#{i-1}]" end

			if f.is_length then vc="(#{vc}).length"
			elsif f.is_data then vc="form[#{f.data_id}]" end

#			if t.respond_to? :xwrite_inline then
#				code << t.xwrite_inline("output",vc) << ";"
#			else
				code<<"#{const[t]}.xwrite #{target},#{vc};"
#			end
		}
		code<<"end"
		@code = code if X11_LOG >= 5
		eval code
		@compiled = true
		#p self,code if self==GraphicsOnDrawable or
		#	self==GraphicsOnDrawable.rem_methods[:fill_rects].in or
		#	self==Graphics.rem_methods[:change].in
	end

	def xread(input)
		args = []
		format.each {|f|
			source=input; source=source.header if f.in_header
			t=f.vtype
			if f.is_regular;   r = t.xread(source); args<<r if f.sym
			elsif f.is_length; args<<t.xread(source)
			elsif f.is_data;   args[f.data_id]=t.xread(input,args[f.data_id])
			end
		}
		self.new(*args)
	end
end

class PartialForm
	def self.xwrite(output,object)
		assert_type "object", object, self if DoCheckTypes
		present = 0
		presenta = []
		s = object.set
		for k,v in s
			i = fields_h[k]
			present |= 1<<i
			presenta[i]=i,v
		end
		presenta.compact!
		Uint32.xwrite output,present
		for i,v in presenta
			fields[i].vtype.xwrite output,v
			output.pad
		end
	end

	def self.xread(input)
		present = Uint32.xread input
		i=0
		s={}
		for i in 0...fields.length
			next unless present[i]>0
			s[fields[i].sym] = fields[i].vtype.xread(input)
		end
		self.new(s)
	end
end

class IntType
	def xwrite(output,object)
		#_{"[self,object]"}
		output.write [object].pack(@packer)
	end
	def xread(input)
		input.read(@length).unpack(@packer)[0]
	end
end

class List
	def xwrite(output,object)
		#X11::FormXWriteUsage[self]||=0
		#X11::FormXWriteUsage[self]+=1
#		xwrite_compile if not @compiled
#		xwrite_compiled(output,object)
		for e in object; @element_type.xwrite output,e; end
		output.pad unless @options == :unpadded
	end

	def xwrite_compile
		p self
		code = "def self.xwrite_compiled(output,object)"
		code << %{assert_type "object", object, self;} if DoCheckTypes
		#!@#$ potential bug for lists of lists. (?)
		if @element_type.respond_to? :xwrite_inline then
			code << %{for e in object; #{@element_type.xwrite_inline("output","e")}; end }
		else
			code << %{for e in object; @element_type.xwrite output,e; end;}
		end
		code << "output.pad;" unless @options == :unpadded
		code << ";end"
		p code
		eval code
		@compiled = true
	end

	def xread(input,count=nil)
		raise "count mismatch" if length and count==length
		count ||= length
		# count ||= chunk.length / l
		raise ItDependsError if not count
		(object = [nil]).pop
		count.times { object << element_type.xread(input) }
		input.pad unless @options == :unpadded
		object
	end
end

class ChoiceType
	attr_accessor :int_type

	alias initialize_old initialize
	def initialize(*a)
		initialize_old(*a)
		@int_type = Uint8
	end		

	def xwrite(output,object)
		int_type.xwrite(output, lookup(object))
	end

	def xread(input)
		v = int_type.xread(input)
		if DoCheckTypes
			self[v].nil? and raise "#{self} has no choice numbered #{v}"
		end
		self[v]
	end
end

class MultiChoiceType
	def xwrite(output,object); int_type.xwrite(output,object); end
	def xread(input);          int_type.xread(input);          end
end

#----------------------------------------------------------------#
# Remote Object Protocol
#----------------------------------------------------------------#

class RequestForm < Form
	class<<self
		attr_accessor :qlass, :rf, :selector, :role
		def to_s; "X11::RequestForm[#{selector.inspect},#{role.inspect}]"; end
	end
end

class RequestFormat < Form; fields_are \
	[:selector, Symbol],
	[:rid, Fixnum],
	[:in, Any.of(Array,Type)], # actually a RequestForm type
	[:out, Any.of(Array,Type,NilClass)] # actually a RequestForm type

	#!@#$ include the class here please
	def initialize(sel,rid,i,o)
		i = make_request_form i, sel, :i if i
		o = make_request_form o, sel, :o if o
		super(sel,rid,i,o)
	end

#	def in;  @in  = make_request_form @in,  sel, :i if Array===@in;  @in;  end
#	def out; @out = make_request_form @out, sel, :o if Array===@out; @out; end

	def make_request_form(arglist,selector,role)
		unless Class===arglist && arglist <= Form
			serf = self
			arglist = Class.new(RequestForm).instance_eval {
				rf = serf
				fields_are *arglist; self
			}
		end
		arglist.role = role
		arglist.selector = selector
		arglist
	end

	def xrequest(xdisplay,receiver,args)
		sid = xdisplay.pick_sid
#		printf "Request 0x%04x\n", sid if xdisplay.sock.log
		stuff = XStringWriter.new(xdisplay)
		Uint8.xwrite stuff.header,rid
		self.in.xwrite(stuff, self.in[*args], receiver)
#		self.in.xwrite(stuff, args, receiver)

		stuff.pad
		len = 1+stuff.string.length/4
		len <= 0x10000 or
			raise TypeError, "request too big (protocol limit: 256k)"
		len <= xdisplay.display_info.maximum_request_length or
			raise TypeError, "request too big (server limit)"

		h = stuff.header
		h.write " " if h.string.length==1

		Uint16.xwrite stuff.header, len
		xdisplay.sock.write(h.string + stuff.string)
#		xdisplay.sock.write(h.string)
#		xdisplay.sock.write(stuff.string)
		return sid
	end
end

# a list of all existing requests (filled at loadtime)
RTable = []

module RemoteClass
	attr_reader :rem_methods

	def def_rem(sym,rid,iargs,oargs)
		@rem_methods ||= {}
		if @rem_methods[sym] then raise "already defined" end

		@rem_methods[sym] = RequestFormat[sym,rid,iargs,oargs]
		module_eval %{def #{sym}(*args,&p); remote_send(:#{sym},args,&p); end}

		RTable[rid] = [self,sym]
	end

	def def_remote(sym,rid,iargs)
		def_rem(sym,rid,iargs,nil)
	end

	alias def_remote_with_reply def_rem
end

module RemoteObject

	attr_reader :xdisplay
	#def xdisplay; @xdisplay or raise "can't find xdisplay in #{self}"; end

	def remote_send(sym,args,&block)
		rform = nil
		self.class.ancestors.each {|t| rform = t.rem_methods[sym]; break if rform }
		rform or raise "remote method not found: sym=#{sym}"
		sid = rform.xrequest(xdisplay,self,args)
		return sid if not rform.out
		done = false
		result = nil
		xdisplay.reply_set[sid&0xffff] = proc {|input|
			result = rform.out.xread(input)
			done = true
			yield result if block
		}
		if block
			wf = @xdisplay.waiting_for
			wf << (sid&0xffff) if wf
			sid
		else
			while not done; xdisplay.xreceive_blocking; end
			result
		end
	end
end

class XID; include RemoteObject; extend RemoteClass, Type
	class<<self
		#!@#$ remember that all remote resources shall have a
		# possible 'closed' status in their local incarnation.
		# this could be XID==-1 or XID==nil.

#		def xwrite(output,object); Uint32.xwrite(output,object.xid); end
		def xwrite(output,object)
			#output.write [object.xid].pack("L")
			output.write(object.xidp||=[object.xid].pack("L"))
		end
		def xread(input); new(input.xdisplay,Uint32.xread(input)); end
	end

	attr_accessor :xid, :xidp

	def initialize(xdisplay,xid=nil)
		assert_type "xdisplay", xdisplay, ::X11::Display if DoCheckTypes
		xid ||= xdisplay.new_id
		assert_type "xid", xid, Uint32 if DoCheckTypes
		@xdisplay,@xid = xdisplay,xid
	end

	def inspect
		#!@#$ should the screen be omitted?
		sprintf "#<%s %08x@%s:%s.%s>",
			self.class,
			@xid,
			@xdisplay.host,
			@xdisplay.display_id,
			@xdisplay.screen_id
	end
end

#----------------------------------------------------------------#
# Events

# an array for number-to-class mapping.
EventTypeNumber = []

class Event < Form
	attr_accessor :sid

	class<<self
		attr_accessor :event_code, :const_name
		def code_is(blah); self.event_code = blah; end
		def event_code=(event_code)
			raise "already there" if @event_code
			@event_code = event_code
			EventTypeNumber[event_code] = self
		end
	end

	def self.xwrite(output,object)
		raise NotImplementedError
	end

	def self.xread(input)
		X11.log1 "Receiving a #{self}"
		chunk = input.read(32)
		chunk = XStringReader.new(input.xdisplay,chunk[4..-1],chunk[0,4])
		chunk.header.read(1)
		obj = super(chunk)
		chunk.header.read(1) if chunk.header.pos==1
		obj.sid = Uint16.xread chunk.header
		obj
	end
end

#----------------------------------------------------------------#
end # module X11
