=begin
	Ruby X11 Client Library
	Copyright 2001 by Mathieu Bouchard

	X11/Remote.rb: socket, auth, i/o.

	$Id: Remote.rb,v 1.4 2002/01/23 03:49:53 matju Exp $

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

=end

require "fcntl"

module X11
#----------------------------------------------------------------#
# Log

X11_LOG = ::ENV["X11_LOG"].to_i

LogColors = {
	:caller => "\e[0;33m",
	:in    => "\e[0;1;35m",
	:event => "\e[0;1;35;44m",
	:out   => "\e[0;1;36m",
	:end   => "\e[0m",
	:success =>"\e[0;1;32m",
	:error => "\e[0;1;31m",
}

def self.log(kind,message)
	STDERR.puts "#{LogColors[kind]}#{message}#{LogColors[:end]}"
end

def self.log1(message)
	STDERR.puts "#{message}" if X11_LOG >= 1
end

#----------------------------------------------------------------#
# NonBlockingSocket
#   this circumvents bugs in versions of IO#read prior to 2001-dec-10.
#   Of course this will cause occasional pauses in execution when using
#   TCP transport.

FAKE_NONBLOCK = true

module NonBlockingSocket
	if FAKE_NONBLOCK then
		def nonblock= flag
			@nonblock = flag
		end
		def read n
			if not @nonblock or IO.select [self],[],[],0 then
				super n
			else
				raise Errno::EWOULDBLOCK
			end
		end
	else # real nonblock (Ruby versions after 2001-dec-10)
		def nonblock= flag
			bit = Fcntl::O_NONBLOCK
			state = fcntl(Fcntl::F_GETFL, 0)
			fcntl(Fcntl::F_SETFL, (state & ~bit) |
				(if flag; bit else 0 end))
		end
	end

	if false then # for future use
		def method_missing sel,*args,&b
			@sock.__send__(sel,*args,&b)
			
		end
	end
end
#----------------------------------------------------------------#
# SocketLog

module SocketLog

	# set this to true to enable logging.
	attr_accessor :log

# category: byte I/O

	# write(data) : writes data
	# read(n)     : reads n bytes exactly (waits for)

	def write(data)
		@count ||= 0
		@count += data.length
		fumble :out, data, caller.join("\n\t") if @log
		super
	end

	def flush
		super
		X11.log :out, "(flush after #{@count} bytes)" if @count>0 && X11_LOG>0
		@count = 0
	end

	def read(n)
		return "" if n==0
		data = super(n)
#		data = ""
#		data << ((super n) || "") while data.length<n
#		raise "short read?" if data.length<n
		fumble :in, data, caller.join("\n\t") if @log
		data
	end

	def hexify(data,prefix1,prefix2)
		data = data.unpack("H*")[0]
		s="#{prefix1}"
		i=0;j=0
		0.step(data.length-1,8) {|x|
			if i==8 then
				i=0;j+=1
				s << if j>=8 then "...\ntotal: #{data.length} bytes.\n"
					else "\\\n#{prefix2}" end
			end
			next if j>=8
			s << data[x,8] << " "
			i+=1
		}
		s
	end

	def fumble(dir,data,foo="")
		# STDOUT.puts "#{ColorCaller}caller=#{foo}\n"
		s=case dir; when :in; "<"; when :out; ">"; end + " "
		s=hexify(data,s,"  ")
		X11.log(dir,s)
	end
end

#----------------------------------------------------------------#

class Host < Form; fields_are \
		[:family, HostFamily], 1,
		[:address, Uint16, :length],
		[:address, List.of(Uint8), :data]
end

# This class is a feature-deprived version of lib/X11/libXau.so
# This is supposed to read stuff from ~/.Xauthority or similar.

class AuthInfo < Form; fields_are \
	[:family,     HostFamilyBE],
	[:address,    Uint16BE, :length], [:address,    String8Unpadded, :data],
	[:display_id, Uint16BE, :length], [:display_id, String8Unpadded, :data],
	[:auth_name,  Uint16BE, :length], [:auth_name,  String8Unpadded, :data],
	[:auth_data,  Uint16BE, :length], [:auth_data,  String8Unpadded, :data]
end

#!@#$ i am old crap code. rewrite me
class Auth
	class << self; alias open new; end

	def initialize(fn=nil)
		@filename =
			fn || ::ENV["XAUTHORITY"] || (::ENV["HOME"]+"/.Xauthority")
		@file = File.open(@filename)
	end

	def read(n)
		buf = @file.read(n)
		raise if buf.length != n
		buf
	end

	def data
		return if @data
		@data=[]
		while not @file.eof
			@data << AuthInfo.xread(XIOReader.new(nil,self))
		end
		@data
	end

	def get_by_host(host,family,display_id)
		if host == "localhost" or host == "127.0.0.1" then
			host = `hostname`
			host.chomp!
		end
		addr = case family
			when :Internet; TCPSocket.gethostbyname(host)
			else nil end
		# puts "host #{host}, addr #{addr}, fam #{family}, dpy #{display_id}"
		data.each {|a|
			#!@#$ if (family==:Internet && a.address==addr or a.address==host)
			if display_id == a.display_id and family == a.family then
				return [a.auth_name,a.auth_data]
			end
		}
		return nil
	end
end

#----------------------------------------------------------------#
end # module X11
